// -*- C++ -*-
//
// HigzManager.h
//
// Author:  Mitsuhiro YAMAGA
// Created: Tue Sep 10 00:52:16 JST 2002
//
// $Id: HigzManager.h,v 1.1 2002/09/27 12:32:30 yamaga Exp $
//
// Revision History
//
// $Log: HigzManager.h,v $
// Revision 1.1  2002/09/27 12:32:30  yamaga
// new package
//
//
//

#ifndef HIGZMANAGER_H_INCLUDED
#define HIGZMANAGER_H_INCLUDED

////////
#include <list>
#include <string>
#include "CLHEP/Vector/ThreeVector.h"
#include "CLHEP/Matrix/Vector.h"
#include "CLHEP/Matrix/Matrix.h"
#include "ccfortran/ccfortran.h"


////////
// object type
enum {
  HIGZ_LINE=1, HIGZ_MARKER, HIGZ_FILL, HIGZ_TEXT
};

// line type
enum {
  LTYP_SOLID=1, LTYP_DASHED, LTYP_DOTTED, LTYP_DASHED_DOTTED
};

// fill style
enum {
  FAIS_HOLLOW=0, FAIS_SOLID, FAIS_PATTERN, FAIS_HATCHED
};

// color
enum{
  WHITE=0,BLACK,RED,GREEN,BLUE,YELLOW,MAGENTA,CYAN
};


//
class  HigzManager {
public:
  // Constructors
  HigzManager( int wkid, int wktype=1 ); // wktype = 1:local, 0:postscript
  ~HigzManager();
  
  // ZEBRA initialization/termination
  static void  init( int nwpawc );
  static void  term();
  
  // Setting up functions
  
  // clear screen
  void    ClearScreen();

  // color
  void    Color( int index_number,double red,double green,double blue );

  // double buffer
  void    DoubleBuffer( bool dblbuf=true );

  // filling style
  void    FillIndex( int index );
  void    FillStyle( int style );

  // hplopt
  void    Hplopt( const std::string& opt );
  void    Hplopt( const std::list<std::string>& opt );

  // igset
  void    Igset( const std::string& opt, double value );
  void    SetDefault( void );
   
  // line attribute
  void    LineType( int type );
  void    LineWidth( double width );

  // Resize Window
  void    ResizeWindow();

  // Rotation
  const HepMatrix&  Rotation( const std::string& axis, // e.g. "zyx"
                              double angle1,double angle2=0,double angle3=0 );
  
  const HepMatrix&  Rotation() const { return m_rot; }
  void  RotationOrigin( const Hep3Vector& org ){ m_rot_orig = org; }

  // Viewport
  void    SetViewport( double xmin,double xmax, double ymin,double ymax );
  void    SetWkViewport( double x0,double x1, double y0,double y1 );
  
  // Window
  void    SetWindow( double xmin,double xmax, double ymin,double ymax );
  void    SetWkWindow( double x0,double x1, double y0,double y1 );

  // Update
  void    Update();
  
  // Update
  void    Zoom( const Hep3Vector& org, double distance );
  
  // Postscript
  void    ClosePsFile();
  void    OpenPsFile( const std::string& filename, int LUN=99 );

  // Drawing functions
  //  in alphabetical order

  // arc
  void    Arc( double xc,double yc,double rmin,double rmax,
	       double phimin,double phimax );
  void    Arc( const Hep3Vector& org, const Hep3Vector& normvect,
	       const Hep3Vector& baseline,
	       double rmin,double rmax,
	       double phimin,double phimax,
	       int nline = 16 );

  // circle
  void    Circle( double xc,double yc,double r );
  void    Circle( const Hep3Vector& org, const Hep3Vector& normvect,
		  double r, int nline = 32 );
  
  // ellipse
  void    Ellipse( double xc,double yc,double rxmin,double rxmax,
		   double rymin,double rymax,double phimin,double phimax,
		   int nline = 32 );
  void    Ellipse( const Hep3Vector& org, 
		   const Hep3Vector& normvect,
		   const Hep3Vector& baseline,
		   double rxmin,double rxmax,
		   double rymin,double rymax,
		   double phimin,double phimax,
		   int nline=32 );

  // key
  void    Key( double x,double y, int sym, const std::string& message );
  void    Key( const Hep3Vector& pos, int sym, const std::string& message );
  
  // line
  void    Line( double x1,double y1, double x2,double y2 );
  void    Line( int npoint,float x[],float y[] );
  void    Line( const Hep3Vector& pt1,const Hep3Vector& pt2  );
  void    Line( int npoint, const Hep3Vector pt[] );
  void    Line( const std::list<Hep3Vector>& pt );

  // marker
  void    Marker( float x,float y,int sym,double size );
  void    Marker( const Hep3Vector& pos,int sym,double size );
 
  // polygon
  void    Polygon( int npoint,float x[],float y[] );
  void    Polygon( int npoint,const Hep3Vector pt[] );

  // puts
  void    Puts( double x,double y,const std::string& message );
  void    Puts( const Hep3Vector& pos, const std::string& message );
  
  // rectangle
  void    Rectangle( double x1,double y1,double x2,double y2 );
  
  //
  //
private:
  //
  static int  s_active_wkid;
  //
  int         m_wkid;
  int         m_wktype;
  int         m_pslun;
  double      m_xmin_wc;
  double      m_xmax_wc;
  double      m_ymin_wc;
  double      m_ymax_wc;
  double      m_zoomdist;
  HepMatrix   m_rot;
  Hep3Vector  m_rot_orig;
  Hep3Vector  m_zoomorg;
  //
  inline void   activate();
  inline Hep3Vector  rotateVector( const Hep3Vector& p );
  inline Hep3Vector  zoomVector( const Hep3Vector& p );
};


// inline
inline void
HigzManager::activate()
{
  if ( s_active_wkid != m_wkid ){
    if ( s_active_wkid > 0 )
      IDAWK( s_active_wkid );
    IACWK( m_wkid );
    ISELNT( m_wkid );
    s_active_wkid = m_wkid;
  }
}

inline Hep3Vector
HigzManager::rotateVector( const Hep3Vector& p )
{
  HepVector  pr(3);
  pr = p - m_rot_orig;
  pr = m_rot * pr;
  return( Hep3Vector( pr(1),pr(2),pr(3) ) + m_rot_orig );
}

inline Hep3Vector
HigzManager::zoomVector( const Hep3Vector& p )
{
  if ( m_zoomdist <= 0 )
    return( p );
  
  Hep3Vector  pz = p - m_zoomorg;
  double  zz = m_zoomdist - pz.z();
  double  sf = (zz > 0) ? m_zoomdist / zz : m_zoomdist * 1.0e-6;
  //cout << p << pz << " " << zz << " " << sf << endl;
  return( pz*sf + m_zoomorg );
}

//
#endif // HIGZMANAGER_H_INCLUDED

