// -*- C++ -*-
//
// HigzManager.cc
//
// Author:  Mitsuhiro YAMAGA
// Created: Tue Sep 10 00:52:16 JST 2002
//
// $Id: HigzManager.cc,v 1.3 2002/10/07 13:46:47 yamaga Exp $
//
// Revision History
//
// $Log: HigzManager.cc,v $
// Revision 1.3  2002/10/07 13:46:47  yamaga
// Added wktype=0 for postscript-only option.
//
// Revision 1.2  2002/09/28 05:26:03  yamaga
// minor bug in parameter list of Rotation()
//
// Revision 1.1  2002/09/27 12:32:30  yamaga
// new package
//
//
//

#include <iostream>
#include <list>
#include <string>
#include <math.h>
#include "CLHEP/Vector/ThreeVector.h"
#include "CLHEP/Matrix/Vector.h"
#include "CLHEP/Matrix/Matrix.h"
#include "ccfortran/ccfortran.h"
#include "HigzManager/HigzManager.h"


////////////////
// MACROS
#define  RAD2DEG (180./M_PI)
#define  DEG2RAD (M_PI/180.)


////////////////
// psfile fortran finctions in HigzPsfile.F
extern"C"{
  void  higz_open_eps_( int* lun, const char* name,int );
  void  higz_open_ps_( int* lun, const char* name,int );
  void  higz_close_ps_( int* lun );
}


////////////////
// static members

////////
int  HigzManager::s_active_wkid = 0;


////////
void
HigzManager::init( int nwpawc )
{
  if ( nwpawc > 0 ){
    MZEBRA( -3 );
    MZPAW( abs(nwpawc),"M" );
  }
  HPLINT( 0 );
}

////////
void
HigzManager::term()
{
  IUWK(0,1);
  HPLEND();
  IGEND();
  s_active_wkid = 0;
}


////////////////
// constructor/destructor

////////
HigzManager::HigzManager( int wkid, int wktype )
  : m_wkid( wkid*10 ),
    m_wktype( wktype ),
    m_pslun( 0 ),
    m_xmin_wc( 0 ),
    m_xmax_wc( 20 ),
    m_ymin_wc( 0 ),
    m_ymax_wc( 20 ),
    m_rot( 3,3,1 ),
    m_rot_orig( 0,0,0 ),
    m_zoomorg( 0,0,0 ),
    m_zoomdist( 0 )
{
  IOPKS( 0 );
  if ( m_wktype > 0 ){
    IOPWK( m_wkid, 0, m_wktype );
  }
  SetWindow( m_xmin_wc, m_xmax_wc, m_ymin_wc, m_ymax_wc );
  SetViewport( m_xmin_wc, m_xmax_wc, m_ymin_wc, m_ymax_wc );
  ISELNT( m_wkid );
  SetDefault();
  //
  //color
  double  rc[] = { 1,0,1,0,0,1,1,0 };
  double  gc[] = { 1,0,0,1,0,1,0,1 };
  double  bc[] = { 1,0,0,0,1,0,1,1 };
  for( int i=0;i<8;i++ ){
    ISCR( m_wkid, i, rc[i],gc[i],bc[i] );
  }
  //
  activate();
}


////////
HigzManager::~HigzManager()
{
  if ( s_active_wkid > 0 ){
    IUWK( m_wkid,1 );
    IDAWK( m_wkid );
    ICLWK( m_wkid );
  }
}


////////////////
// Setting up functions
  
// clear screen
void
HigzManager::ClearScreen()
{
  ICLRWK( m_wkid,1 );
}

// color
void
HigzManager::Color( int index_number,double red,double green,double blue )
{
  ISCR( m_wkid,index_number,red,green,blue );
}


// double buffer
void
HigzManager::DoubleBuffer( bool dblbuf )
{
  activate();
  Igset( "2BUF", (dblbuf) ? m_wkid*10+1 : 0 );
}


// filling style
void
HigzManager::FillIndex( int index )
{
  activate();
  ISFASI( index );
}


void
HigzManager::FillStyle( int style )
{
  activate();
  ISFAIS( style );
}


// opt
void
HigzManager::Hplopt( const std::string& opt )
{
  activate();
  HPLOPT( opt );
}


void
HigzManager::Hplopt( const std::list<std::string>& opt )
{
  activate();
  HPLOPT( opt );
}


// igset
void
HigzManager::Igset( const std::string& opt, double value )
{
  activate();
  IGSET( opt,value );
}


void
HigzManager::SetDefault( void )
{
  activate();
  IGSET( "*",0 );
}


// line attribute
void
HigzManager::LineType( int type )
{
  activate();
  ISLN( type );
}


void
HigzManager::LineWidth( double width )
{
  activate();
  ISLWSC( width );
}


// Resize Window
void
HigzManager::ResizeWindow()
{
  float  f[2];
  IGQWK( m_wkid,"2BUF",f );
  DoubleBuffer( false );
  if ( f[0] > 0.5 ){
    IGQWK( m_wkid,"MXDS",f );
    SetWkViewport( 0,f[0],0,f[1] );
    DoubleBuffer();
  }
  else{
    IGQWK( m_wkid,"MXDS",f );
    SetWkViewport( 0,f[0],0,f[1] );
  }
}



// line attribute
const HepMatrix&
HigzManager::Rotation( const std::string& axis,
                       double angle1,double angle2,double angle3 )
{
  m_rot = HepMatrix( 3,3,1 );
  for( int i=0; i<3 && i<axis.size(); i++ ){
    double  ang = 0;
    switch ( i ){
    case 0 : ang = angle1 * DEG2RAD; break;
    case 1 : ang = angle2 * DEG2RAD; break;
    case 2 : ang = angle3 * DEG2RAD; break;
    };
    HepMatrix  r(3,3,1);
    double  cs = cos( ang );
    double  sn = sin( ang );
    switch ( axis.c_str()[i] ){
    case 'x' :
      r(2,2) =  cs;
      r(2,3) = -sn;
      r(3,2) =  sn;
      r(3,3) =  cs;
      m_rot = r * m_rot;
      break;
    case 'y' :
      r(3,3) =  cs;
      r(3,1) = -sn;
      r(1,3) =  sn;
      r(1,1) =  cs;
      m_rot = r * m_rot;
      break;
    case 'z' :
      r(1,1) =  cs;
      r(1,2) = -sn;
      r(2,1) =  sn;
      r(2,2) =  cs;
      m_rot = r * m_rot;
      break;
    default :
      std::cout << " HigzManager::Rotation : "
                << " axis name invalid : " << axis.c_str()[i]
                << std::endl;
    }
  }
  return( m_rot );
}


// Viewport
void
HigzManager::SetViewport( double xmin,double xmax, double ymin,double ymax )
{
  double  x0 = (xmin - m_xmin_wc) / (m_xmax_wc - m_xmin_wc);
  double  x1 = (xmax - m_xmin_wc) / (m_xmax_wc - m_xmin_wc);
  double  y0 = (ymin - m_ymin_wc) / (m_ymax_wc - m_ymin_wc);
  double  y1 = (ymax - m_ymin_wc) / (m_ymax_wc - m_ymin_wc);
  ISVP( m_wkid,x0,x1,y0,y1 );
  ISELNT( m_wkid );
}


// Workstation Viewport
void
HigzManager::SetWkViewport( double x0,double x1, double y0,double y1 )
{
  ISWKVP( m_wkid,x0,x1,y0,y1 );
  ISELNT( m_wkid );
}

 
// Window
void
HigzManager::SetWindow( double xmin,double xmax, double ymin,double ymax )
{
  m_xmin_wc = xmin;
  m_xmax_wc = xmax;
  m_ymin_wc = ymin;
  m_ymax_wc = ymax;
  ISWN( m_wkid,xmin,xmax,ymin,ymax );
  ISELNT( m_wkid );
}


// Workstation Window
void
HigzManager::SetWkWindow( double x0,double x1, double y0,double y1 )
{
  ISWKWN( m_wkid,x0,x1,y0,y1 );
  ISELNT( m_wkid );
}

 
// Update
void
HigzManager::Update()
{
  IUWK( m_wkid,1 );
}


// Zoom
void
HigzManager::Zoom( const Hep3Vector& org, double distance )
{
  m_zoomorg = org;
  m_zoomdist = distance;
}


// Postscript
void
HigzManager::ClosePsFile()
{
  higz_close_ps_( &m_pslun );
}

void
HigzManager::OpenPsFile( const std::string& filename, int lun )
{
  m_pslun = lun;
  if ( filename.find( ".eps" ) == filename.size() - 4 ){
    higz_open_eps_( &m_pslun, filename.c_str(), filename.size() );
  }
  else{
    higz_open_ps_( &m_pslun, filename.c_str(), filename.size() );
  }
}


// Drawing functions
//  in alphabetical order

// arc
void
HigzManager::Arc( double xc,double yc,double rmin,double rmax,
                  double phimin,double phimax )
{
  activate();
  IGARC( xc,yc,rmin,rmax,phimin,phimax );
}


void
HigzManager::Arc( const Hep3Vector& org, const Hep3Vector& normvect,
                  const Hep3Vector& baseline,
                  double rmin,double rmax,
                  double phimin,double phimax,
                  int nline )
{
  activate();

  //  nline
  if ( nline < 1 )
    nline = 1;
  
  // var.
  Hep3Vector*  pt = new Hep3Vector[nline*2+2];

  // normalize
  Hep3Vector  norm  = normvect.unit();
  Hep3Vector  trans = norm.cross( baseline ).unit();
  Hep3Vector  base  = trans.cross( norm ).unit();
  
  // phi
  double  dphi = (phimax - phimin) / nline ;
  // inner arc
  int  ct = 0;
  for( int i=0;i<=nline;i++ ){
    double  phi = (phimin + dphi*i)*DEG2RAD;
    pt[ct++] = org + base*rmin*cos( phi ) + trans*rmin*sin( phi );
  }
  // outer arc
  for( int i=nline;i>=0;i-- ){
    double  phi = (phimin + dphi*i)*DEG2RAD;
    pt[ct++] = org + base*rmax*cos( phi ) + trans*rmax*sin( phi );
  }
  //  show
  Polygon( ct,pt );

  // free
  delete [] pt;
}



// circle
void
HigzManager::Circle( double xc,double yc,double r )
{
  activate();

  IGARC( xc,yc,r,r,0,360 );
}


void
HigzManager::Circle( const Hep3Vector& org, const Hep3Vector& normvect,
                     double r, int nline )
{
  activate();

  // nline
  if ( nline < 1 )
    nline = 1;
  
  // var.
  Hep3Vector*  pt = new Hep3Vector[nline+1];

  // normalize
  Hep3Vector  norm = normvect.unit();
  Hep3Vector  base = Hep3Vector( 1.0, 0.0, 0.0 );
  if ( norm == base ){
    base = Hep3Vector( 0.0, 1.0, 0.0 );
  }
  Hep3Vector  trans = norm.cross( base ).unit();
  base = trans.cross( norm ).unit();
  
  // angle
  double  dphi = 2.0*M_PI / nline;
  
  // Polygon
  int  ct = 0;
  for( int i=0;i<nline;i++ ){
    double  phi = dphi * i;
    pt[ct++] = org + base * r * cos( phi ) + trans * r * sin( phi );
  }
  Polygon( ct,pt );
  
  // free
  delete [] pt;
}

  
// ellipse
void
HigzManager::Ellipse( double xc,double yc,double rxmin,double rxmax,
                      double rymin,double rymax,double phimin,double phimax,
                      int nline )
{
  activate();

  Ellipse( Hep3Vector(xc,yc,0),Hep3Vector(0,0,1),Hep3Vector(1,0,0),
           rxmin, rxmax, rymin, rymax, phimin, phimax, nline );
}


void
HigzManager::Ellipse( const Hep3Vector& org, 
                      const Hep3Vector& normvect,
                      const Hep3Vector& baseline,
                      double rxmin,double rxmax,
                      double rymin,double rymax,
                      double phimin,double phimax,
                      int nline )
{
  activate();

  //  nline
  if ( nline < 1 )
    nline = 1;

  // var.
  Hep3Vector*  pt = new Hep3Vector [nline*2+2];
    
  // normalize
  Hep3Vector  norm  = normvect.unit();
  Hep3Vector  trans = norm.cross( baseline ).unit();
  Hep3Vector  base  = trans.cross( norm ).unit();

  // phi
  double  dphi = (phimax - phimin) / nline ;

  // inner arc
  int  ct = 0;
  if ( rxmin > 0 && rymin > 0 ){
    for( int i=0;i<=nline;i++ ){
      double  phi = (phimin + dphi * i) * DEG2RAD;
      double  cs = cos(phi);
      double  sn = sin(phi);
      double  r = 1.0 / sqrt( cs * cs / rxmin / rxmin +
                              sn * sn / rymin / rymin );
      pt[ct++] = org + base * r * cs + trans * r * sn;
    }
  }
  // outer arc
  if ( rxmax > 0 && rymax > 0 ){
    for( int i=nline;i>=0;i-- ){
      double  phi = (phimin + dphi * i) * DEG2RAD;
      double  cs = cos(phi);
      double  sn = sin(phi);
      double  r = 1.0 / sqrt( cs * cs / rxmax / rxmax +
                              sn * sn / rymax / rymax );
      pt[ct++] = org + base * r * cs + trans * r * sn;
    }
  }
  //  show
  Polygon( ct,pt );

  // free
  delete [] pt;
}


// key
void
HigzManager::Key( double x,double y, int sym, const std::string& message )
{
  activate();

  HPLKEY( x,y,sym,message );
}


void
HigzManager::Key( const Hep3Vector& pos, int sym, const std::string& message )
{
  activate();
  
  Hep3Vector  p = zoomVector( rotateVector( pos ) );
  HPLKEY( p.x(),p.y(), sym, message );
}

  
// line
void
HigzManager::Line( double x1,double y1, double x2,double y2 )
{
  activate();

  float   x[2] = {x1,x2};
  float   y[2] = {y1,y2};
  int     n=2;
  
  IPL( n,x,y );
}


void
HigzManager::Line( int npoint,float x[],float y[] )
{
  activate();
  IPL( npoint,x,y );
}


void
HigzManager::Line( const Hep3Vector& pt1,const Hep3Vector& pt2  )
{
  activate();
  Hep3Vector  p1 = zoomVector( rotateVector( pt1 ) );
  Hep3Vector  p2 = zoomVector( rotateVector( pt2 ) );
  Line( p1.x(),p1.y(), p2.x(),p2.y() );
}


void
HigzManager::Line( int npoint,const Hep3Vector pt[] )
{
  activate();

  // nline
  if ( npoint < 1 )
    npoint = 1;
  
  // var
  float*  x = new float[npoint];
  float*  y = new float[npoint];

  // coordinate
  for( int i=0;i<npoint;i++ ){
    Hep3Vector  p = zoomVector( rotateVector( pt[i] ) );
    x[i] = p.x();
    y[i] = p.y();
  }
  
  // show
  IPL( npoint,x,y );
  
  // free
  delete y;
  delete x;
}


void
HigzManager::Line( const std::list<Hep3Vector>& pt )
{
  activate();

  // var
  int  n = pt.size();
  float*  x = new float[n];
  float*  y = new float[n];

  // coordinate
  int  ct = 0;
  for( std::list<Hep3Vector>::const_iterator i=pt.begin();
       i!=pt.end(); i++ ){
    Hep3Vector  p = zoomVector( rotateVector( *i ) );
    x[ct] = p.x();
    y[ct] = p.y();
    ct++;
  }
  
  // show
  IPL( n,x,y );
  
  // free
  delete y;
  delete x;
}


// marker
void
HigzManager::Marker( float x, float y, int sym, double size )
{
  activate();

  ISMKSC( size );
  ISMK( sym );
  IPM( 1,&x,&y );
}


void
HigzManager::Marker( const Hep3Vector& pos,int sym,double size )
{
  activate();

  ISMKSC( size );
  ISMK( sym );
  Hep3Vector  p = zoomVector( rotateVector( pos ) );
  float  fx = p.x();
  float  fy = p.y();
  IPM( 1, &fx, &fy );
}

 
// polygon
void
HigzManager::Polygon( int npoint,float x[],float y[] )
{
  activate();
  IFA( npoint,x,y );
}

void
HigzManager::Polygon( int npoint,const Hep3Vector pt[] )
{
  activate();

  // nline
  if ( npoint < 1 )
    npoint = 1;
  
  // var
  float*  x = new float [ npoint ];
  float*  y = new float [ npoint ];
  
  for( int i=0;i<npoint;i++ ){
    Hep3Vector  p = zoomVector( rotateVector( pt[i] ) );
    x[i] = p.x();
    y[i] = p.y();
  }

  IFA( npoint,x,y );

  // delete
  delete y;
  delete x;
}


// puts
void
HigzManager::Puts( double x_wc, double y_wc, const std::string& message )
{
  activate();

  ITX( x_wc, y_wc, message );
}


void
HigzManager::Puts( const Hep3Vector& pos, const std::string& message )
{
  activate();

  Hep3Vector  p = zoomVector( rotateVector( pos ) );
  ITX( p.x(), p.y(), message );
}


// rectangle
void
HigzManager::Rectangle( double x1,double y1,double x2,double y2 )
{
  activate();

  float       x[4],y[4];
  x[0] = x1;    y[0] = y1;
  x[1] = x2;    y[1] = y1;
  x[2] = x2;    y[2] = y2;
  x[3] = x1;    y[3] = y2;
  Polygon( 4,x,y );
}

